module bwi_search_mod
  use base_struct_mod
  use bwi_kernel_mod
  use bwi_restom_mod
  implicit none

contains

  integer function bwi_sdstep_oper(srch_, grdo_, vgrid) result(stat)
    character(len=*)                    :: srch_, grdo_
    type(vel_grid)                      :: vgrid
    real, dimension(:,:,:), allocatable :: srch

    allocate(srch(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
    call bwi_fio_vmod_updt(srch, vgrid, grdo_, "reed")
    srch = -srch
    call bwi_fio_vmod_updt(srch, vgrid, srch_, "rite")
    deallocate(srch)
    stat = 0.
  end function

  integer function bwi_cgstep_oper(srch_, grdo_, grdp_, vgrid, restart) result(stat)
    character(len=*)                    :: srch_, grdo_, grdp_
    type(vel_grid)                      :: vgrid
    logical                             :: restart

    double precision                    :: beta
    real, dimension(:,:,:), allocatable :: grdo
    real, dimension(:,:,:), allocatable :: grdp
    real, dimension(:,:,:), allocatable :: srch

    allocate(grdo(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
    allocate(grdp(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
    allocate(srch(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))

!   gradient for the k'th iteration
    call bwi_fio_vmod_updt(grdo, vgrid, grdo_, "reed") ! k
!   gradient for the k+1'th iteration
    call bwi_fio_vmod_updt(grdp, vgrid, grdp_, "reed") ! k+1
!   restarting
    if (restart) then
       beta = 0.
    else
!      compute beta, using the Polak-Ribiere method
       beta = sum(dprod(grdp,grdp-grdo))/(sum(dprod(grdo,grdo))+epsilon(beta))
!      safe guarding
       beta = max(beta, 0.)
    end if
!    write(0,*)"beta = ", beta
!   read the search direction for the k'th iteration (current)
    call bwi_fio_vmod_updt(srch, vgrid, srch_, "reed")
!   update the search direction
    srch = - grdo + beta*srch
!   write out new search direction
    call bwi_fio_vmod_updt(srch, vgrid, srch_, "rite")
!   shuffle the gradient
    call bwi_fio_vmod_updt(grdp, vgrid, grdo_, "rite")
    deallocate(grdo)
    deallocate(grdp)
    deallocate(srch)
    stat = 0
  end function
 
  integer function bwi_linsec_oper(csou_, crec_, bvel_, bimg_, hess_, srch_, imsk_, vgrid, igrid, hgrid, nsrc, alph, fobj, &
                                   hcut, dsotyp, dsorms, objtyp, objeps, grdfix) result(stat)
    character(len=*)                    :: csou_, crec_, bvel_, bimg_, hess_, srch_, imsk_
    type(vel_grid)                      :: vgrid
    type(img_grid)                      :: igrid
    type(hes_grid)                      :: hgrid
    integer                             :: nsrc
    real                                :: alph
    double precision                    :: fobj
    real                                :: hcut
    integer                             :: dsotyp
    logical                             :: dsorms
    character(len=6)                    :: objtyp
    real                                :: objeps
    logical                             :: grdfix

    double precision                    :: f(nsrc), fopt, norm
    real                                :: sfac
    integer                             :: iopt, isrc

    real, dimension(:,:,:), allocatable :: grdo
    real, dimension(:,:,:), allocatable :: grdp
    real, dimension(:,:,:), allocatable :: srch

    allocate(grdo(vgrid%ax%n,vgrid%ay%n,vgrid%az%n)) 
    allocate(grdp(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
    allocate(srch(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))

    sfac = 0.5*(nsrc+1)

!   read the background model to be updated
!   ---------------------------------------
    call bwi_fio_vmod_updt(grdo, vgrid, bvel_, "reed")
!   read the search direction
!   -------------------------
    call bwi_fio_vmod_updt(srch, vgrid, srch_, "reed")
!   normalize
!   ---------
    norm = maxval(abs(srch))
    srch = srch/norm
!   perturb the model
!   -----------------
    do isrc=1, nsrc
!      trial velocity model
!      -------------------- 
       grdp = grdo + alph*isrc*srch
       call bwi_fio_vmod_updt(grdp, vgrid, bvel_, "rite")
!      compute the image using the trial velocity
!      ------------------------------------------
       stat = bwi_kerimg_oper(.true., .false., csou_, crec_, bvel_, bimg_)
!      compute the objective function
!      ------------------------------
       stat = bwi_restom_oper(bimg_, hess_, imsk_, igrid, hgrid, f(isrc), hcut, dsotyp, dsorms, objtyp, objeps, grdfix)
    end do
!   find the minimum one
!   --------------------
    fopt = f(1)
    iopt = 1
    do isrc=1, nsrc
       if (f(isrc)<=fopt) then
          iopt = isrc
          fopt = f(isrc)
       end if
    end do
!   compare with the current objective function value
!   -------------------------------------------------
    if (f(iopt)<=fobj) then
       grdp = grdo + alph*iopt*srch
       alph = alph*iopt/sfac
       fobj = f(iopt)
    else
       grdp = grdo
       alph = alph * 0.5
    end if
    call bwi_fio_vmod_updt(grdp, vgrid, bvel_, "rite")
    deallocate(grdo)
    deallocate(grdp)
    deallocate(srch)
    stat = 0
  end function


end module
